import attrs
from ..core import Field
from ..core.mixin import WithQualifiers
from ..core.exceptions import FormatMismatchError


@attrs.define
class Singluar(Field):
    pass


class LogicalMixin:
    def __bool__(self):
        return bool(self.value)

    def __and__(self, other):
        return self.value and other.value

    def __or__(self, other):
        return self.value or other.value

    def __not__(self):
        return not self.value

    def __invert__(self):
        return not self.value


class ScalarMixin(LogicalMixin):
    def __add__(self, other):
        return self.value + other

    def __sub__(self, other):
        return self.value - other

    def __mul__(self, other):
        return self.value * other

    def __truediv__(self, other):
        return self.value / other

    def __floordiv__(self, other):
        return self.value // other

    def __mod__(self, other):
        return self.value % other

    def __pow__(self, other):
        return self.value**other

    def __neg__(self):
        return -self.value

    def __pos__(self):
        return +self.value

    def __abs__(self):
        return abs(self.value)


def text_converter(value):
    try:
        return str(value)
    except ValueError as e:
        raise FormatMismatchError(str(e))


def integer_converter(value):
    if isinstance(value, float):
        raise FormatMismatchError(
            f"Cannot convert float ({value}) to integer field without potential loss "
            "of information"
        )
    try:
        return int(value)
    except ValueError as e:
        raise FormatMismatchError(str(e))


def decimal_converter(value):
    try:
        return float(value)
    except ValueError as e:
        raise FormatMismatchError(str(e))


def boolean_converter(value):
    if isinstance(value, str):
        if value.lower() in ("true", "1", "yes"):
            value = True
        elif value.lower() in ("false", "0", "no"):
            value = False
        else:
            raise FormatMismatchError(
                f"Cannot convert string '{value}' to boolean value"
            )
    else:
        try:
            value = bool(value)
        except ValueError as e:
            raise FormatMismatchError(str(e))
    return value


def array_converter(value):
    if isinstance(value, str):
        if value.startswith("[") and value.endswith("]"):
            value = value[1:-1]
        elif value.startswith("[") or value.endswith("]"):
            raise FormatMismatchError(f"Unmatched brackets in array field {value}")
        value = [v.strip() for v in value.split(",")]
    else:
        try:
            value = list(value)
        except ValueError as e:
            raise FormatMismatchError(str(e))
    return value


@attrs.define
class Text(Singluar):

    value: str = attrs.field(converter=text_converter)

    raw_type = str


@attrs.define
class Integer(Singluar, ScalarMixin):

    value: int = attrs.field(converter=integer_converter)

    raw_type = int

    def __int__(self):
        return self.value


@attrs.define
class Decimal(Singluar, ScalarMixin):

    value: float = attrs.field(converter=decimal_converter)

    raw_type = int

    def __float__(self):
        return self.value


@attrs.define
class Boolean(Singluar, LogicalMixin):

    raw_type = bool

    value: bool = attrs.field(converter=boolean_converter)

    def __str__(self):
        return str(self.value).lower()

    def __bool__(self):
        return self.value


@attrs.define
class Array(WithQualifiers, Field):

    # WithQualifiers class attrs
    qualifiers_attr_name = "item_type"
    multiple_qualifiers = False
    allowed_qualifiers = (Singluar,)
    item_type = None

    raw_type = list

    value: list = attrs.field(converter=array_converter)

    def __attrs_post_init__(self):
        # Ensure items are of the correct type
        if self.item_type is not None:
            self.value = [self.item_type(i).value for i in self.value]

    def __str__(self):
        return (
            "["
            + ",".join(
                str(self.item_type(i)) if self.item_type is not None else i
                for i in self.value
            )
            + "]"
        )

    def __iter__(self):
        return iter(self.value)



from pathlib import Path
import json
from ..core import __version__, mark, DataType
from ..core.mixin import WithQualifiers
from ..generic import File
from ..core.utils import MissingExtendedDependency

try:
    import yaml
except ImportError:
    yaml = MissingExtendedDependency("yaml", __name__)


class Schema(DataType):
    """Base class for all serialization schemas (can be used for abstract schemas that
    don't actually have a formal definition)"""

    pass


class JsonSchema(Schema):
    pass


class XmlSchema(Schema):
    pass


class InformalSchema(Schema):
    "Not actually a strict schema, just a set of conventions on how to structure the serialization"


class DataSerialization(WithQualifiers, File):
    "Base class for text-based hierarchical data-serialization formats, e.g. JSON, YAML"

    # Qualifiers class attrs
    qualifiers_attr_name = "schema"
    schema = None
    multiple_qualifiers = False
    allowed_qualifiers = (Schema,)
    generically_qualifies = True

    iana_mime = None

    @mark.check
    def load(self):
        """Load the contents of the file into a dictionary"""
        raise NotImplementedError

    @classmethod
    def save_new(fspath: Path, dct: dict):
        """Serialise a dictionary to a new file"""
        raise NotImplementedError


class Xml(DataSerialization):
    ext = ".xml"
    allowed_qualifiers = (XmlSchema, InformalSchema)


class Json(DataSerialization):
    ext = ".json"
    allowed_qualifiers = (JsonSchema, InformalSchema)

    @mark.check
    def load(self):
        with open(self.fspath) as f:
            dct = json.load(f)
        return dct

    @classmethod
    def save_new(cls, fspath, dct):
        with open(fspath, "w") as f:
            json.dump(dct, f)
        return cls(fspath)


class Yaml(DataSerialization):
    ext = ".yaml"
    alternate_exts = (".yml",)

    @mark.check
    def load(self):
        with open(self.fspath) as f:
            dct = yaml.load(f, Loader=yaml.Loader)
        return dct

    @classmethod
    def save_new(cls, fspath, dct):
        with open(fspath, "w") as f:
            yaml.dump(dct, f)
        return cls(fspath)



import typing as ty
import attrs
from .utils import describe_task
from .exceptions import FileFormatsError


@attrs.define
class ConverterWrapper:
    """Wraps a converter task in a workflow so that the in_file and out_file names can
    be mapped onto their standardised names, "in_file" and "out_file" if necessary
    """

    task_spec: ty.Callable
    in_file: str = None
    out_file: str = None

    def __call__(self, name=None, **kwargs):
        from pydra.engine import Workflow

        if name is None:
            name = f"{self.task_spec.__name__}_wrapper"
        wf = Workflow(
            name=name, input_spec=list(set(["in_file"] + list(kwargs))), **kwargs
        )
        wf.add(self.task_spec(name="task", **{self.in_file: wf.lzin.in_file}))
        wf.set_output([("out_file", getattr(wf.task.lzout, self.out_file))])
        return wf


class SubtypeVar:
    """To handle the case where the target format is a placeholder (type-var) defined by
    by its relationship to the source format, e.g.

    AnyFileFormat = FileSet.type_var("AnyFileFormat")

    @converter
    @pydra.mark.task
    def unzip(in_file: Zip[AnyFileFormat], out_file: AnyFileFormat):
        ...
    """

    converters = {}

    def __init__(self, name: str, base: type):
        self.name = name
        self.base = base

    def __str__(self):
        return f"{self.base.__name__}:{self.name}"

    def __repr__(self):
        return f"SubtypeVar({str(self)})"

    def __hash__(self):
        return hash((type(self), self.base, self.name))

    @property
    def __name__(self):
        return self.name

    def issubtype(self, super_type: type, allow_same: bool = True):
        """Check to see whether datatype class is a subtype of a given super class.
        In this case the subtype is expected to be able to be treated as if it was
        the super class.

        Overridden in the ``WithQualifiers`` mixin to add support for
        qualified subtypes

        Parameters
        ----------
        super_type : type
            the class to check whether the given class is a subtype of
        allow_same : bool, optional
            whether there is a match if the classes are the same, by default True

        Returns
        -------
        is_subtype : bool
            whether or not the current class can be considered a subtype of the super (or
            is the super itself)
        """
        return self.base.issubtype(super_type, allow_same=allow_same)

    @classmethod
    def get_converter_tuples(
        cls, source_format: type, target_format: type
    ) -> ty.List[ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]]]:
        # check to see whether there are converters from a base class of the source
        # format
        available_converters = []
        if source_format.is_qualified:
            for template_source_format, converter in cls.converters.items():
                if not template_source_format.unqualified.issubtype(
                    source_format.unqualified
                ):
                    continue
                assert len(template_source_format.wildcard_qualifiers()) == 1
                non_wildcards = template_source_format.non_wildcard_qualifiers()
                if not non_wildcards.issubset(source_format.qualifiers):
                    continue
                from_types = tuple(
                    set(source_format.qualifiers).difference(non_wildcards)
                )
                if any(q.issubtype(target_format) for q in from_types):
                    available_converters.append(converter)
        return available_converters

    @classmethod
    def register_converter(
        cls,
        source_format: type,
        converter_tuple: ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]],
    ):
        """Registers a converter task within a class attribute. Called by the @fileformats.mark.converter
        decorator.

        Parameters
        ----------
        source_format : type
            the source format to register a converter from
        task_spec : ty.Callable
            a callable that resolves to a Pydra task
        converter_kwargs : dict
            additional keyword arguments to be passed to the task spec at initialisation
            time

        Raises
        ------
        FormatConversionError
            if there is already a converter registered between the two types
        """
        # Ensure "converters" dict is defined in the target class and not in a superclass
        if len(source_format.wildcard_qualifiers()) > 1:
            raise FileFormatsError(
                "Cannot register a conversion to a generic type from a type with more "
                f"than one wildcard {source_format} ({list(source_format.wildcard_qualifiers())})"
            )
        prev_registered = [
            f
            for f in cls.converters
            if (
                f.unqualified is source_format.unqualified
                and f.non_wildcard_qualifiers()
                == source_format.non_wildcard_qualifiers()
            )
        ]
        assert len(prev_registered) <= 1
        if prev_registered:
            prev = prev_registered[0]
            prev_task = cls.converters[prev][0]
            raise FileFormatsError(
                f"There is already a converter registered from {prev} "
                f"to the generic type '{tuple(prev.wildcard_qualifiers())[0]}':"
                f"{describe_task(prev_task)}"
            )

        cls.converters[source_format] = converter_tuple



from ..core import __version__
from fileformats.generic import File
from fileformats.archive import Zip
from fileformats.core.mixin import WithMagicNumber


# Document formats
class Document(File):
    iana_mime = None


class Pdf(WithMagicNumber, Document):
    ext = ".pdf"
    magic_number = b"%PDF-"
    binary = True

    iana_mime = "application/pdf"


class Msword(Document):
    ext = ".doc"
    binary = True

    iana_mime = "application/msword"


class MswordX(Zip, Document):
    ext = ".docx"

    iana_mime = (
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )


class Postscript(WithMagicNumber, Document):
    ext = ".eps"
    alternate_exts = (".ps",)
    magic_number = b"%!"
    iana_mime = "application/postscript"



import pytest
from fileformats.generic import File, Directory, DirectoryContaining
from fileformats.core.exceptions import FormatMismatchError
from fileformats.core.mixin import WithSideCars
from conftest import write_test_file


def test_generic_file(work_dir):

    fspath = work_dir / "test.txt"
    write_test_file(fspath)
    file = File(fspath)
    assert file.fspath == fspath
    assert str(file) == str(fspath)


def test_generic_file_fail(work_dir):

    fspath = work_dir / "test-dir"
    fspath.mkdir()
    with pytest.raises(FormatMismatchError):
        File(fspath)


def test_generic_dir(work_dir):

    fspath = work_dir / "test-dir"
    fspath.mkdir()
    d = Directory(fspath)
    assert d.fspath == fspath
    assert str(d) == str(fspath)


def test_generic_dir_fail(work_dir):

    fspath = work_dir / "test.txt"
    write_test_file(fspath)
    with pytest.raises(FormatMismatchError):
        Directory(fspath)


def test_dynamic_dir(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.foo")
    assert DirectoryContaining[TestFile].matches(fspath)


def test_dynamic_dir_fail(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.bad")
    assert not DirectoryContaining[TestFile].matches(fspath)


class TestFile(File):

    ext = ".foo"


def test_file_ext(work_dir):

    fspath = work_dir / "test.foo"
    write_test_file(fspath)
    assert TestFile.matches(fspath)


def test_file_ext_fail(work_dir):

    fspath = work_dir / "test.bad"
    write_test_file(fspath)
    assert not TestFile.matches(fspath)


def test_single_of_double_ext(work_dir):

    fspath = work_dir / "test.bar.foo"
    write_test_file(fspath)
    assert TestFile.matches(fspath)


def test_file_cast(work_dir):

    fspath = work_dir / "test.foo"
    write_test_file(fspath)
    file = TestFile(fspath)
    assert File(file).fspath == fspath


class TestDir(Directory):

    content_types = (TestFile,)


def test_dir_contents(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.foo")
    write_test_file(fspath / "test2.foo")
    write_test_file(fspath / "text.txt")
    assert TestDir.matches(fspath)


def test_dir_contents_fail(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.bad")
    write_test_file(fspath / "test.zip")
    write_test_file(fspath / "text.txt")
    assert not TestDir.matches(fspath)


class DoubleExtFileFormat(File):

    ext = ".foo.bar"


def test_double_ext(work_dir):
    fspath = work_dir / "test.foo.bar"
    write_test_file(fspath)
    assert DoubleExtFileFormat.matches(fspath)


def test_double_ext_fail(work_dir):
    fspath = work_dir / "test.foo.bad"
    write_test_file(fspath)
    assert not DoubleExtFileFormat.matches(fspath)


class NestedDirFormat(Directory):

    content_types = (TestFile, DoubleExtFileFormat, TestDir)


def test_nested_directories(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.foo")
    write_test_file(fspath / "text.ignored")
    write_test_file(fspath / "text.foo.bar")
    write_test_file(fspath / "nested_dir" / "test.foo")
    write_test_file(fspath / "nested_dir" / "test.ignored")
    assert NestedDirFormat.matches(fspath)


def test_nested_directories_fail(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.foo")
    write_test_file(fspath / "text.foo.bar")
    write_test_file(fspath / "nested_dir" / "test.bad")
    write_test_file(fspath / "nested_dir" / "test.ignored")
    assert not NestedDirFormat.matches(fspath)


def test_nested_directories_fail2(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.foo")
    write_test_file(fspath / "text.foo.bad")
    write_test_file(fspath / "nested_dir" / "test.foo")
    write_test_file(fspath / "nested_dir" / "test.ignored")
    assert not NestedDirFormat.matches(fspath)


# class FileWithSideCars(File):

#     ext = ".foo"

#     @mark.required
#     @property
#     def bar(self):
#         return self.select_by_ext(".bar")


# def test_side_car(work_dir):
#     fspath = work_dir / "test.foo"
#     write_test_file(fspath)
#     bar_fspath = work_dir / "test.bar"
#     write_test_file(bar_fspath)
#     file = FileWithSideCars([fspath, bar_fspath])
#     assert file.bar == bar_fspath


# def test_side_car2(work_dir):
#     fspath = work_dir / "test.foo"
#     write_test_file(fspath)
#     bar_fspath = work_dir / "whoopwhoop.bar"
#     write_test_file(bar_fspath)
#     file = FileWithSideCars([fspath, bar_fspath])
#     assert file.fspath == fspath
#     assert file.bar == bar_fspath


# def test_side_car_from_primary(work_dir):
#     fspath = work_dir / "test.foo"
#     write_test_file(fspath)
#     bar_fspath = work_dir / "test.bar"
#     write_test_file(bar_fspath)
#     file = FileWithSideCars(fspath)
#     assert file.fspath == fspath
#     assert file.bar == bar_fspath


# def test_side_car_fail(work_dir):
#     fspath = work_dir / "test.foo"
#     write_test_file(fspath)
#     assert not FileWithSideCars.matches(fspath)


# def test_side_car_fail2(work_dir):
#     fspath = work_dir / "test.foo"
#     write_test_file(fspath)
#     fspath = work_dir / "test.bad"
#     write_test_file(fspath)
#     assert not FileWithSideCars.matches(fspath)


class Bar(File):

    ext = ".bar"


class FooWithSideCars(WithSideCars, TestFile):

    side_car_types = (Bar,)


class DirContainingSideCars(Directory):

    content_types = (FooWithSideCars,)


def test_dir_containing_side_cars(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.foo")
    write_test_file(fspath / "test.bar")
    write_test_file(fspath / "text.txt")
    assert DirContainingSideCars.matches(fspath)


def test_dir_containing_side_cars_fail(work_dir):

    fspath = work_dir / "test-dir"
    write_test_file(fspath / "test.foo")
    write_test_file(fspath / "diff-name.bar")
    assert not DirContainingSideCars.matches(fspath)



import importlib
from pathlib import Path
import inspect
import typing as ty
import re
import os
import pkgutil
from contextlib import contextmanager
from fileformats.core.exceptions import (
    MissingExtendedDepenciesError,
    FileFormatsError,
)
import fileformats.core


def find_matching(fspaths: ty.List[Path], standard_only: bool = False):
    """Detect the corresponding file format from a set of file-system paths

    Parameters
    ----------
    fspaths : list[Path]
        file-system paths to detect the format of
    standard_only : bool, optional
        If you only want to return matches from the "standard" IANA types
    """
    fspaths = fspaths_converter(fspaths)
    matches = []
    for frmt in fileformats.core.FileSet.all_formats:
        if frmt.matches(fspaths) and (
            not standard_only or frmt.namespace in STANDARD_NAMESPACES
        ):
            matches.append(frmt)
    return matches


def from_mime(mime_str: str):
    return fileformats.core.DataType.from_mime(mime_str)


def splitext(fspath: Path, multi=False):
    """splits an extension from the file stem, taking into consideration multi-part
    extensions such as ".nii.gz".

    Parameters
    ----------
    fspath : Path
        the file-system path to split the extension from
    multi : bool, optional
        whether to support multi-part extensions such as ".nii.gz". Note this means that
        it will match sections of file names with "."s in them, by default False

    Returns
    -------
    str
        file stem
    str
        file extension
    """
    if multi:
        ext = "".join(fspath.suffixes)
        stem = fspath.name[: -len(ext)]
    else:
        stem = fspath.stem
        ext = fspath.suffix
    return stem, ext


def subpackages():
    """Iterates over all subpackages within the fileformats namespace

    Yields
    ------
    module
        all modules within the package
    """
    for mod_info in pkgutil.iter_modules(
        fileformats.__path__, prefix=fileformats.__package__ + "."
    ):
        if mod_info.name == "core":
            continue
        yield importlib.import_module(mod_info.name)


@contextmanager
def set_cwd(path: Path):
    """Sets the current working directory to `path` and back to original
    working directory on exit

    Parameters
    ----------
    path : str
        The file system path to set as the current working directory
    """
    pwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(pwd)


def fspaths_converter(
    fspaths: ty.Union[
        ty.Iterable[ty.Union[str, os.PathLike, bytes]], str, os.PathLike, bytes
    ]
):
    """Ensures fs-paths are a set of pathlib.Path"""
    if isinstance(fspaths, fileformats.core.FileSet):
        fspaths = fspaths.fspaths
    elif isinstance(fspaths, (str, os.PathLike, bytes)):
        fspaths = [fspaths]
    return frozenset(Path(p).absolute() for p in fspaths)


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class MissingExtendedDependency:
    """Used as a placeholder for package dependencies that are only installed with the
    "extended" install extra.

    Parameters
    ----------
    pkg_name : str
        name of the package to act as a placeholder for
    module_path : str
        path of the module, i.e. the value of the '__name__' global variable
    """

    def __init__(self, pkg_name: str, module_path: str):
        self.pkg_name = pkg_name
        module_parts = module_path.split(".")
        assert module_parts[0] == "fileformats"
        if module_parts[1] in STANDARD_NAMESPACES:
            self.required_in = "fileformats"
        else:
            self.required_in = f"fileformats-{module_parts[1]}"

    def __getattr__(self, _):
        raise MissingExtendedDepenciesError(
            f"Not able to access extended behaviour in {self.required_in} as required "
            f"package '{self.pkg_name}' was not installed. Please reinstall "
            f"{self.required_in} with 'extended' install extra to access extended "
            f"behaviour of the format classes (such as loading and conversion), i.e.\n\n"
            f"    $ python3 -m pip install {self.required_in}[extended]"
        )


STANDARD_NAMESPACES = [
    "archive",
    "audio",
    "document",
    "image",
    "numeric",
    "serialization",
    "text",
    "video",
]


def to_mime_format_name(format_name: str):
    if "___" in format_name:
        raise FileFormatsError(
            f"Cannot convert name of format class {format_name} to mime string as it "
            "contains triple underscore"
        )
    format_name = format_name[0].lower() + format_name[1:]
    format_name = re.sub("__([A-Z])", lambda m: "+" + m.group(1).lower(), format_name)
    format_name = re.sub("_([A-Z])", lambda m: "." + m.group(1).lower(), format_name)
    format_name = re.sub("([A-Z])", lambda m: "-" + m.group(1).lower(), format_name)
    return format_name


def from_mime_format_name(format_name: str):
    if format_name.startswith("x-"):
        format_name = format_name[2:]
    format_name = format_name.capitalize()
    format_name = re.sub(r"(\.)(\w)", lambda m: "_" + m.group(2).upper(), format_name)
    format_name = re.sub(r"(\+)(\w)", lambda m: "__" + m.group(2).upper(), format_name)
    format_name = re.sub(r"(-)(\w)", lambda m: m.group(2).upper(), format_name)
    return format_name


def hash_file(fspath: Path, chunk_len: int, crypto: ty.Callable):
    crypto_obj = crypto()
    with open(fspath, "rb") as fp:
        for chunk in iter(lambda: fp.read(chunk_len), b""):
            crypto_obj.update(chunk)
    return crypto_obj.hexdigest()


def hash_dir(
    fspath: Path,
    chunk_len: int,
    crypto: ty.Callable,
    ignore_hidden_files: bool = False,
    ignore_hidden_dirs: bool = False,
    relative_to: Path = None,
):
    if relative_to is None:
        relative_to = fspath
    file_hashes = {}
    for dpath, _, filenames in sorted(os.walk(fspath)):
        # Sort in-place to guarantee order.
        filenames.sort()
        dpath = Path(dpath)
        if ignore_hidden_dirs and dpath.name.startswith(".") and str(dpath) != fspath:
            continue
        for filename in filenames:
            if ignore_hidden_files and filename.startswith("."):
                continue
            file_hashes[str((dpath / filename).relative_to(relative_to))] = hash_file(
                dpath / filename, crypto=crypto, chunk_len=chunk_len
            )
    return file_hashes


def add_exc_note(e, note):
    """Adds a note to an exception in a Python <3.11 compatible way

    Parameters
    ----------
    e : Exception
        the exception to add the note to
    note : str
        the note to add

    Returns
    -------
    Exception
        returns the exception again
    """
    if hasattr(e, "add_note"):
        e.add_note(note)
    else:
        e.args = (e.args[0] + "\n" + note,)
    return e


def describe_task(task):
    """Returns the name of a Pydra task and where it was defined for debugging purposes

    Parameters
    ----------
    task : pydra.engine.core.TaskBase
        the task to describe
    """
    from fileformats.core.converter import ConverterWrapper

    if isinstance(task, ConverterWrapper):
        task = task.task_spec
    if inspect.isfunction(task):
        import cloudpickle

        task = cloudpickle.loads(task().inputs._func)
    src_file = inspect.getsourcefile(task)
    src_line = inspect.getsourcelines(task)[-1]
    return f"{task} (defined at line {src_line} of {src_file})"



from conftest import write_test_file
from fileformats.generic import File
from fileformats.core import mark
from fileformats.core.mixin import WithMagicNumber, WithSeparateHeader, WithSideCars
from fileformats.core.exceptions import FormatMismatchError


class FileWithMagicNumber(WithMagicNumber, File):

    ext = ".magic"
    binary = True
    magic_number = b"magicnumber"


def test_magic_number(work_dir):

    fspath = work_dir / "test.magic"
    write_test_file(
        fspath,
        FileWithMagicNumber.magic_number + b"some contents\n\n",
        binary=True,
    )
    assert FileWithMagicNumber.matches(fspath)


def test_magic_fail(work_dir):

    fspath = work_dir / "test.magic"
    write_test_file(fspath, b"NOMAGIC some contents\n\n", binary=True)
    assert not FileWithMagicNumber.matches(fspath)


class Header(File):

    ext = ".hdr"

    def load(self):
        return dict(ln.split(":") for ln in self.contents.splitlines())


class FileWithSeparateHeader(WithSeparateHeader, File):

    ext = ".img"
    header_type = Header

    image_type_key = "image-type"
    image_type = "sample-image-type"

    @mark.check
    def check_image_type(self):
        if self.metadata[self.image_type_key] != self.image_type:
            raise FormatMismatchError(
                f"Mismatch in '{self.image_type_key}', expected {self.image_type}, "
                f"found {self.metadata[self.image_type_key]}"
            )


def test_with_separate_header(work_dir):

    fspath = work_dir / "image.img"
    write_test_file(fspath)
    hdr_fspath = work_dir / "image.hdr"
    hdr = {
        FileWithSeparateHeader.image_type_key: FileWithSeparateHeader.image_type,
        "dims": "10,10,20",
    }
    write_test_file(hdr_fspath, "\n".join(f"{k}:{v}" for k, v in hdr.items()))
    file = FileWithSeparateHeader(fspath)
    assert file.metadata == hdr


def test_with_separate_header_fail1(work_dir):

    fspath = work_dir / "image.img"
    write_test_file(fspath)
    hdr_fspath = work_dir / "image.bad"
    hdr = {
        FileWithSeparateHeader.image_type_key: FileWithSeparateHeader.image_type,
        "dims": "10,10,20",
    }
    write_test_file(hdr_fspath, "\n".join(f"{k}:{v}" for k, v in hdr.items()))
    assert not FileWithSeparateHeader.matches([fspath, hdr_fspath])


def test_with_separate_header_fail2(work_dir):

    fspath = work_dir / "image.img"
    write_test_file(fspath)
    hdr_fspath = work_dir / "image.hdr"
    hdr = {
        FileWithSeparateHeader.image_type_key: "bad-type",
        "dims": "10,10,20",
    }
    write_test_file(hdr_fspath, "\n".join(f"{k}:{v}" for k, v in hdr.items()))
    assert not FileWithSeparateHeader.matches([fspath, hdr_fspath])


class ImageWithInlineHeader(File):

    ext = ".img"
    binary = True

    image_type_key = "image-type"
    image_type = "sample-image-type"

    header_separator = b"---END HEADER---"

    def load_metadata(self):
        hdr = self.contents.split(self.header_separator)[0].decode("utf-8")
        return dict(ln.split(":") for ln in hdr.splitlines())


class FileWithSideCars(WithSideCars, ImageWithInlineHeader):

    ext = ".img"
    primary_type = ImageWithInlineHeader
    side_car_types = (Header,)

    experiment_type_key = "experiment-type"
    experiment_type = "sample-experiment-type"

    @mark.check
    def check_image_type(self):
        """Loaded from inline-header"""
        if self.metadata[self.image_type_key] != self.image_type:
            raise FormatMismatchError(
                f"Mismatch in '{self.image_type_key}', expected {self.image_type}, "
                f"found {self.metadata[self.image_type_key]}"
            )

    @mark.check
    def check_experiment_type(self):
        """Loaded from side-car"""
        if self.metadata["header"][self.experiment_type_key] != self.experiment_type:
            raise FormatMismatchError(
                f"Mismatch in '{self.experiment_type_key}', expected {self.experiment_type}, "
                f"found {self.metadata['header'][self.experiment_type_key]}"
            )


def test_with_side_car(work_dir):

    fspath = work_dir / "image.img"
    hdr = {
        ImageWithInlineHeader.image_type_key: ImageWithInlineHeader.image_type,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    side_car_fs_path = work_dir / "image.hdr"
    side_car = {
        FileWithSideCars.experiment_type_key: FileWithSideCars.experiment_type,
    }
    write_test_file(
        side_car_fs_path, "\n".join(f"{k}:{v}" for k, v in side_car.items())
    )
    assert FileWithSideCars.matches([fspath, side_car_fs_path])


def test_with_side_car_fail1(work_dir):

    fspath = work_dir / "image.img"
    hdr = {
        ImageWithInlineHeader.image_type_key: ImageWithInlineHeader.image_type,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + b"BAD SEPARATOR"  # NB: bad separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    side_car_fs_path = work_dir / "image.hdr"
    side_car = {
        FileWithSideCars.experiment_type_key: FileWithSideCars.experiment_type,
    }
    write_test_file(
        side_car_fs_path, "\n".join(f"{k}:{v}" for k, v in side_car.items())
    )
    assert not FileWithSideCars.matches([fspath, side_car_fs_path])


def test_with_side_car_fail2(work_dir):

    fspath = work_dir / "image.img"
    hdr = {
        ImageWithInlineHeader.image_type_key: "bad-image-type",  # NB: bad type
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    side_car_fs_path = work_dir / "image.hdr"
    side_car = {
        FileWithSideCars.experiment_type_key: FileWithSideCars.experiment_type,
    }
    write_test_file(
        side_car_fs_path, "\n".join(f"{k}:{v}" for k, v in side_car.items())
    )
    assert not FileWithSideCars.matches([fspath, side_car_fs_path])


def test_with_side_car_fail3(work_dir):

    fspath = work_dir / "image.img"
    hdr = {
        ImageWithInlineHeader.image_type_key: ImageWithInlineHeader.image_type,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    side_car_fs_path = work_dir / "image.hdr"
    side_car = {
        FileWithSideCars.experiment_type_key: "bad-experiment-type",  # NB: bad type
    }
    write_test_file(
        side_car_fs_path, "\n".join(f"{k}:{v}" for k, v in side_car.items())
    )
    assert not FileWithSideCars.matches([fspath, side_car_fs_path])



from ..core import __version__
from fileformats.generic import File


class Video(File):
    "Base class for audio file formats"
    binary = True
    iana_mime = None


class Mp4(Video):
    ext = ".mp4"
    iana_mime = "video/mp4"


class Webm(Video):
    ext = ".webm"
    iana_mime = "video/webm"


class Quicktime(Video):
    ext = ".mov"
    alternate_exts = (".qt",)
    iana_mime = "video/quicktime"


class Ogg(Video):
    ext = ".ogv"
    iana_mime = "video/ogg"



from fileformats.generic import FileSet
from fileformats.core.utils import from_mime


def test_mime_roundtrip():

    for klass in FileSet.all_formats:
        mimetype = klass.mime_type
        assert isinstance(mimetype, str)
        reloaded = from_mime(mimetype)
        assert reloaded is klass


def test_mimelike_roundtrip():

    for klass in FileSet.all_formats:
        mimetype = klass.mime_like
        assert isinstance(mimetype, str)
        reloaded = from_mime(mimetype)
        assert reloaded is klass



import os
from pathlib import Path
import attrs
from ..core.base import FileSet
from ..core.exceptions import FormatMismatchError, FileFormatsError
from ..core import mark
from ..core.utils import splitext, classproperty
from ..core.mixin import WithQualifiers


@attrs.define
class FsObject(FileSet, os.PathLike):
    "Generic file-system object, can be either a file or a directory"

    @mark.required
    @property
    def fspath(self):
        if len(self.fspaths) > 1:
            raise FormatMismatchError(
                f"More than one fspath ({self.fspaths}) provided to {self}, "
                f"primary path is ambiguous"
            )
        return next(iter(self.fspaths))

    def __str__(self):
        return str(self.fspath)

    def __fspath__(self):
        """Render to string, so can be treated as any other file-system path, i.e. passed
        to functions like file 'open'"""
        return str(self)


@attrs.define
class File(FsObject):
    """Generic file type"""

    binary = False
    is_dir = False

    @mark.required
    @property
    def fspath(self):
        fspath = self.select_by_ext()
        if not fspath.is_file():
            if not fspath.exists():
                msg = (
                    f"Selected fspath for {type(self)}, '{str(fspath)}', does not exist"
                )
                if fspath.parent.exists():
                    neighbours = [p.name for p in fspath.parent.iterdir()]
                    msg += f". Found neighbouring files {neighbours}"
                raise FileFormatsError(msg)
            else:
                msg = ", it is a directory" if fspath.is_dir() else ""
                raise FormatMismatchError(
                    f'Path that matches extension of "{type(self)}", {fspath}, is not a '
                    "file in {repr(self)}" + msg
                )
        return fspath

    @classmethod
    def copy_ext(cls, old_path: Path, new_path: Path):
        """Copy extension from the old path to the new path, ensuring that all
        of the extension is used (e.g. 'my.gz' instead of 'gz')

        Parameters
        ----------
        old_path: Path or str
            The path from which to copy the extension from
        new_path: Path or str
            The path to append the extension to

        Returns
        -------
        Path
            The new path with the copied extension
        """
        if not cls.matching_exts([old_path], [cls.ext]):
            raise FormatMismatchError(
                f"Extension of old path ('{str(old_path)}') does not match that "
                f"of file, '{cls.ext}'"
            )
        suffix = cls.ext if cls.ext else splitext(old_path, multi=True)[-1]
        return Path(new_path).with_suffix(suffix)

    @property
    def contents(self):
        return self.read_contents()

    def read_contents(self, size=None, offset=0):
        with open(self.fspath, "rb" if self.binary else "r") as f:
            if offset:
                f.read(offset)
            contents = f.read(size)
        return contents

    @classproperty
    def possible_exts(cls):
        possible = [cls.ext]
        try:
            possible.extend(cls.alternate_exts)
        except AttributeError:
            pass
        return possible

    @property
    def actual_ext(self):
        "The actual file extension (out of the primary  and alternate extensions possible)"
        matching = [e for e in self.possible_exts if self.fspath.name.endswith(e)]
        if not matching:
            assert False, (
                f"extension of fspath {self.fspath} is not in possible extensions for "
                f"{type(self)} class: {self.possible_exts}. This should have been "
                "checked at initialisation"
            )
        # Return the longest matching extension, useful for optional extensions
        return sorted(matching, key=lambda x: len(x))[-1]

    @property
    def stem(self):
        return self.fspath.name[: -len(self.actual_ext)]


@attrs.define
class Directory(FsObject):
    """Base directory to be overridden by subtypes that represent directories but don't
    want to inherit content type "qualifers" (i.e. most of them)"""

    is_dir = True

    content_types = ()

    @mark.required
    @property
    def fspath(self):
        dirs = [p for p in self.fspaths if p.is_dir()]
        if not dirs:
            raise FormatMismatchError(f"No directory paths provided {repr(self)}")
        elif len(dirs) > 1:
            raise FormatMismatchError(
                f"More than one directory path provided {dirs} to {repr(self)}"
            )
        fspath = dirs[0]
        missing = []
        for content_type in self.content_types:
            match = False
            for p in fspath.iterdir():
                try:
                    content_type([p])
                except FormatMismatchError:
                    continue
                else:
                    match = True
                    break
            if not match:
                missing.append(content_type)
        if missing:
            raise FormatMismatchError(
                f"Did not find matches for {missing} content types in {repr(self)}"
            )
        return fspath

    @property
    def contents(self):
        for content_type in self.content_types:
            for p in self.fspath.iterdir():
                try:
                    yield content_type([p])
                except FormatMismatchError:
                    continue

    @mark.check
    def validate_contents(self):
        if not self.content_types:
            return
        not_found = set(self.content_types)
        for fspath in self.fspath.iterdir():
            for content_type in list(not_found):
                if content_type.matches(fspath):
                    not_found.remove(content_type)
                    if not not_found:
                        return
        assert not_found
        raise FormatMismatchError(
            f"Did not find the required content types, {not_found}, within the "
            f"directory {self.fspath} of {self}"
        )

    def hash_files(self, relative_to=None, **kwargs):
        if relative_to is None:
            relative_to = self.fspath
        return super().hash_files(relative_to=relative_to, **kwargs)


class TypedSet(FileSet):
    """List of specific file types (similar to the contents of a directory but not
    enclosed in one)"""

    content_types = ()

    @property
    def contents(self):
        for content_type in self.content_types:
            for p in self.fspaths:
                try:
                    yield content_type([p])
                except FormatMismatchError:
                    continue

    @mark.check
    def validate_contents(self):
        if not self.content_types:
            return
        not_found = set(self.content_types)
        for fspath in self.fspaths:
            for content_type in list(not_found):
                if content_type.matches(fspath):
                    not_found.remove(content_type)
                    if not not_found:
                        return
        assert not_found
        raise FormatMismatchError(
            f"Did not find the required content types, {not_found}, within the "
            f"given list {self.fspaths}"
        )


class DirectoryContaining(WithQualifiers, Directory):
    """Generic directory qualified by the formats of its contents"""

    # WithQualifiers-required class attrs
    qualifiers_attr_name = "content_types"
    allowed_qualifiers = (FileSet,)
    generically_qualifies = True


class SetOf(WithQualifiers, TypedSet):

    # WithQualifiers-required class attrs
    qualifiers_attr_name = "content_types"
    allowed_qualifiers = (FileSet,)
    generically_qualifies = True



import os.path
import sys
import typing as ty
import tempfile
import tarfile
import zipfile
from pathlib import Path
import attrs
import pydra.mark
import pydra.engine.specs
from fileformats.generic import FsObject
from fileformats.core.utils import set_cwd
from fileformats.core import mark, FileSet
from fileformats.archive import Zip, Tar, TarGzip


TAR_COMPRESSION_TYPES = ["", "gz", "bz2", "xz"]

TAR_COMPRESSION_ANNOT = (
    str,
    {
        "help_string": (
            "The type of compression applied to tar file, "
            "', '".join(TAR_COMPRESSION_TYPES)
        ),
        "allowed_values": list(TAR_COMPRESSION_TYPES),
    },
)

ZIP_COMPRESSION_ANNOT = (
    int,
    {
        "help_string": (
            "The type of compression applied to zip file, "
            "see https://docs.python.org/3/library/zipfile.html#zipfile.ZIP_DEFLATED "
            "for valid compression types"
        ),
        "allowed_values": [
            zipfile.ZIP_STORED,
            zipfile.ZIP_DEFLATED,
            zipfile.ZIP_BZIP2,
            zipfile.ZIP_LZMA,
        ],
    },
)

Compressed = FileSet.type_var("Compressed")

PathLike = ty.Union[str, bytes, os.PathLike]


@mark.converter(source_format=FsObject, target_format=Tar)
@mark.converter(source_format=FsObject, target_format=TarGzip, compression="gz")
@mark.converter(source_format=Compressed, target_format=Tar[Compressed])
@mark.converter(
    source_format=Compressed, target_format=TarGzip[Compressed], compression="gz"
)
@pydra.mark.task
@pydra.mark.annotate(
    {
        "in_file": PathLike,
        "out_file": str,
        "filter": str,
        "compression": str,  # TAR_COMPRESSION_ANNOT,
        "format": int,
        "ignore_zeros": bool,
        "return": {"out_file": Path},
    }
)
def create_tar(
    in_file,
    out_file=None,
    base_dir=None,
    filter=None,
    compression=None,
    format=tarfile.DEFAULT_FORMAT,
    ignore_zeros=False,
    encoding=tarfile.ENCODING,
):
    return _create_tar(
        in_file=in_file,
        out_file=out_file,
        base_dir=base_dir,
        filter=filter,
        compression=compression,
        format=format,
        ignore_zeros=ignore_zeros,
        encoding=encoding,
    )


# @mark.converter(source_format=Directory, target_format=Tar_Gzip, compression="gz")
# @mark.converter(source_format=Directory, target_format=Tar)
# @pydra.mark.task
# @pydra.mark.annotate(
#     {
#         "in_file": pydra.engine.specs.Directory,
#         "out_file": str,
#         "filter": str,
#         "compression": str,  # TAR_COMPRESSION_ANNOT,
#         "format": int,
#         "ignore_zeros": bool,
#         "return": {"out_file": Path},
#     }
# )
# def tar_dir(
#     in_file,
#     out_file=None,
#     base_dir=None,
#     filter=None,
#     compression=None,
#     format=tarfile.DEFAULT_FORMAT,
#     ignore_zeros=False,
#     encoding=tarfile.ENCODING,
# ):
#     return _create_tar(
#         in_file=in_file,
#         out_file=out_file,
#         base_dir=base_dir,
#         filter=filter,
#         compression=compression,
#         format=format,
#         ignore_zeros=ignore_zeros,
#         encoding=encoding,
#     )


def _create_tar(
    in_file,
    out_file=None,
    base_dir=None,
    filter=None,
    compression=None,
    format=tarfile.DEFAULT_FORMAT,
    ignore_zeros=False,
    encoding=tarfile.ENCODING,
):

    if not compression:
        compression = ""
        ext = ".tar"
    else:
        ext = ".tar." + compression

    if not out_file:
        out_file = Path(in_file).name + ext

    if base_dir is None:
        base_dir = Path(in_file).parent

    out_file = os.path.abspath(out_file)

    with tarfile.open(
        out_file,
        mode=f"w:{compression}",
        format=format,
        ignore_zeros=ignore_zeros,
        encoding=encoding,
    ) as tfile, set_cwd(base_dir):
        for fspath in in_file.fspaths:
            tfile.add(relative_path(fspath, base_dir), filter=filter)

    return Path(out_file)


@mark.converter(source_format=Tar, target_format=FsObject)
@mark.converter(source_format=TarGzip, target_format=FsObject)
@mark.converter(source_format=Tar[Compressed], target_format=Compressed)
@mark.converter(source_format=TarGzip[Compressed], target_format=Compressed)
@pydra.mark.task
@pydra.mark.annotate({"return": {"out_file": pydra.engine.specs.MultiOutputObj}})
def extract_tar(
    in_file: PathLike,
    extract_dir: PathLike,
    bufsize: int = 10240,
    compression_type: str = "*",
) -> ty.Iterable[Path]:

    if extract_dir == attrs.NOTHING:
        extract_dir = tempfile.mkdtemp()
    else:
        extract_dir = os.path.abspath(extract_dir)
        os.makedirs(extract_dir, exist_ok=True)
    extract_dir = Path(extract_dir)

    if not compression_type:
        compression_type = ""

    with tarfile.open(in_file, mode=f"r:{compression_type}") as tfile:
        tfile.extractall(path=extract_dir)

    return [extract_dir / f for f in os.listdir(extract_dir)]


@mark.converter(source_format=FsObject, target_format=Zip)
@mark.converter(source_format=Compressed, target_format=Zip[Compressed])
@pydra.mark.task
@pydra.mark.annotate(
    {
        "in_file": PathLike,
        "out_file": str,
        "compression": int,  # ZIP_COMPRESSION_ANNOT,
        "allowZip64": bool,
        "return": {"out_file": PathLike},
    }
)
def create_zip(
    in_file,
    out_file,
    base_dir,
    compression=zipfile.ZIP_DEFLATED,
    allowZip64=True,
    compresslevel=None,
    strict_timestamps=True,
):
    return _create_zip(
        in_file=in_file,
        out_file=out_file,
        base_dir=base_dir,
        compression=compression,
        allowZip64=allowZip64,
        compresslevel=compresslevel,
        strict_timestamps=strict_timestamps,
    )


# @mark.converter(source_format=Directory, target_format=Zip)
# @pydra.mark.task
# @pydra.mark.annotate(
#     {
#         "in_file": pydra.engine.specs.Directory,
#         "out_file": str,
#         "compression": int,  # ZIP_COMPRESSION_ANNOT,
#         "allowZip64": bool,
#         "return": {"out_file": Path},
#     }
# )
# def zip_dir(
#     in_file,
#     out_file,
#     base_dir,
#     compression=zipfile.ZIP_DEFLATED,
#     allowZip64=True,
#     compresslevel=None,
#     strict_timestamps=True,
# ):
#     return _create_zip(
#         in_file=in_file,
#         out_file=out_file,
#         base_dir=base_dir,
#         compression=compression,
#         allowZip64=allowZip64,
#         compresslevel=compresslevel,
#         strict_timestamps=strict_timestamps,
#     )


def _create_zip(
    in_file,
    out_file,
    base_dir,
    compression=zipfile.ZIP_DEFLATED,
    allowZip64=True,
    compresslevel=None,
    strict_timestamps=True,
):

    if out_file == attrs.NOTHING:
        out_file = Path(in_file).name + ".zip"

    if base_dir == attrs.NOTHING:
        base_dir = Path(in_file).parent

    out_file = os.path.abspath(out_file)

    zip_kwargs = {}
    if not strict_timestamps:  # Truthy is the default in earlier versions
        if sys.version_info.major <= 3 and sys.version_info.minor < 8:
            raise Exception(
                "Must be using Python >= 3.8 to pass "
                f"strict_timestamps={strict_timestamps!r}"
            )

        zip_kwargs["strict_timestamps"] = strict_timestamps

    with zipfile.ZipFile(
        out_file,
        mode="w",
        compression=compression,
        allowZip64=allowZip64,
        compresslevel=compresslevel,
        **zip_kwargs,
    ) as zfile, set_cwd(base_dir):
        for fspath in in_file.fspaths:
            fspath = Path(fspath)
            if fspath.is_dir():
                for dpath, _, files in os.walk(fspath):
                    zfile.write(relative_path(dpath, base_dir))
                    for fname in files:
                        fpath = os.path.join(dpath, fname)
                        zfile.write(relative_path(fpath, base_dir))
            else:
                zfile.write(relative_path(fspath, base_dir))
    return Path(out_file)


@mark.converter(source_format=Zip, target_format=FsObject)
@mark.converter(source_format=Zip[Compressed], target_format=Compressed)
@pydra.mark.task
@pydra.mark.annotate({"return": {"out_file": pydra.engine.specs.MultiOutputObj}})
def extract_zip(in_file: PathLike, extract_dir: PathLike):

    if extract_dir == attrs.NOTHING:
        extract_dir = tempfile.mkdtemp()
    else:
        extract_dir = os.path.abspath(extract_dir)
        os.makedirs(extract_dir, exist_ok=True)
    extract_dir = Path(extract_dir)

    with zipfile.ZipFile(in_file) as zfile:
        zfile.extractall(path=extract_dir)

    return [extract_dir / f for f in os.listdir(extract_dir)]


def relative_path(path, base_dir):
    path = os.path.abspath(path)
    relpath = os.path.relpath(path, base_dir)
    if ".." in relpath:
        raise RuntimeError(
            f"Cannot add {path} to archive as it is not a "
            f"subdirectory of {base_dir}"
        )
    return relpath



from ..core import __version__
from fileformats.generic import File


class TextMatrix(File):
    ext = ".mat"


class RFile(File):
    ext = ".rData"
    binary = True


class MatlabMatrix(File):
    ext = ".mat"


class DataFile(File):
    """Generic binary data file"""

    binary = True
    ext = ".dat"



import pytest
from fileformats.generic import File
from fileformats.core.mixin import WithSeparateHeader
from fileformats.core.utils import (
    find_matching,
    MissingExtendedDependency,
)
from fileformats.core.exceptions import MissingExtendedDepenciesError
import fileformats.text
import fileformats.numeric
from conftest import write_test_file


class Bar(File):

    ext = ".bar"


class Foo(WithSeparateHeader, File):

    ext = ".foo"
    header_type = Bar


def test_copy(work_dir):

    foo_fspath = work_dir / "x.foo"
    write_test_file(foo_fspath)
    bar_fspath = work_dir / "y.bar"
    write_test_file(bar_fspath)
    file = Foo([foo_fspath, bar_fspath])
    dest_dir = work_dir / "new-dir"
    dest_dir.mkdir()
    cpy = file.copy_to(dest_dir)
    assert cpy.fspath.parent == dest_dir
    assert cpy.fspath.name == "x.foo"
    assert cpy.header.fspath.parent == dest_dir
    assert cpy.header.fspath.name == "y.bar"


def test_copy_with_rename(work_dir):

    foo_fspath = work_dir / "x.foo"
    write_test_file(foo_fspath)
    bar_fspath = work_dir / "y.bar"
    write_test_file(bar_fspath)
    file = Foo([foo_fspath, bar_fspath])
    cpy = file.copy_to(work_dir, stem="new")
    assert cpy.fspath.name == "new.foo"
    assert cpy.header.fspath.name == "new.bar"


def test_copy_symlink(work_dir):

    foo_fspath = work_dir / "x.foo"
    write_test_file(foo_fspath)
    bar_fspath = work_dir / "y.bar"
    write_test_file(bar_fspath)
    file = Foo([foo_fspath, bar_fspath])
    dest_dir = work_dir / "new-dir"
    dest_dir.mkdir()
    cpy = file.copy_to(dest_dir, symlink=True)
    assert cpy.fspath.parent == dest_dir
    assert cpy.fspath.name == "x.foo"
    assert cpy.header.fspath.parent == dest_dir
    assert cpy.header.fspath.name == "y.bar"


def test_copy_ext(work_dir):
    assert (
        File.copy_ext(work_dir / "x.foo.bar", work_dir / "y") == work_dir / "y.foo.bar"
    )
    assert Bar.copy_ext(work_dir / "x.foo.bar", work_dir / "y") == work_dir / "y.bar"


def test_format_detection(work_dir):

    text_file = work_dir / "text.txt"

    with open(text_file, "w") as f:
        f.write("sample text")

    detected = find_matching(text_file, standard_only=True)
    assert len(detected) == 1
    assert detected[0] is fileformats.text.Plain


def test_missing_dependency():

    missing_dep = MissingExtendedDependency("missing_dep", "fileformats.image")

    with pytest.raises(MissingExtendedDepenciesError):
        missing_dep.an_attr



from ..core import __version__
from fileformats.generic import File
from fileformats.serialization import (
    Json,
    Xml,
    Yaml,
)  # These are sometimes/historically considered part of the text registry so we import them here


# General formats
class Plain(File):
    ext = ".txt"
    iana_mime = "text/plain"


class Csv(File):
    ext = ".csv"
    iana_mime = "text/csv"


class Tsv(File):
    ext = ".tsv"
    iana_mime = "text/tsv"


class Html(File):
    ext = ".html"
    alternate_exts = (".htm",)
    iana_mime = "text/html"


class Markdown(File):
    ext = ".md"
    iana_mime = "text/markdown"


class RestructedText(File):
    ext = ".rst"
    iana_mime = "text/x-rst"



class FileFormatsError(RuntimeError):
    "Base exception class"


class FormatMismatchError(RuntimeError):
    "File formats don't match"


class FormatConversionError(RuntimeError):
    "No converters exist between formats"


class FormatRecognitionError(KeyError):
    "Did not find a format class corresponding to a MIME, or MIME-like, type string"


class MissingExtendedDepenciesError(RuntimeError):
    "'extended' install extra wasn't installed required for advanced behaviour"



import filecmp
import pytest
from fileformats.generic import File, Directory
from fileformats.archive import Zip, Gzip, Tar, TarGzip
from fileformats.text import Plain as PlainText


TEST_DIR = "__test_dir__"
TEST_FILE = "__test_file__.txt"

INPUT_TYPES = ["file", "directory"]


@pytest.fixture(params=INPUT_TYPES)
def archive_input(work_dir, request):
    if request.param == "file":
        test_file = work_dir / TEST_FILE
        with open(test_file, "w") as f:
            f.write("test file contents")
        inpt = PlainText(test_file)
    elif request.param == "directory":
        test_dir = work_dir / TEST_DIR
        test_dir.mkdir()
        for i in range(1, 3):
            with open(test_dir / f"file{i}.txt", "w") as f:
                f.write(f"test file {i}")
        sub_dir = test_dir / "sub-dir"
        sub_dir.mkdir()
        for i in range(4, 6):
            with open(sub_dir / f"file{i}.txt", "w") as f:
                f.write(f"test file {i}")
        inpt = Directory(test_dir)
    else:
        assert False, f"Unrecognised request param {request.param}"
    return inpt


def test_zip_roundtrip(archive_input):
    _roundtrip(archive_input, Zip)


@pytest.mark.xfail(reason="Gzip converter is not implemented yet")
def test_gzip_roundtrip(archive_input):
    _roundtrip(archive_input, Gzip)


def test_tar_roundtrip(archive_input):
    _roundtrip(archive_input, Tar)


def test_tar_gz_roundtrip(archive_input):
    _roundtrip(archive_input, TarGzip)


def _roundtrip(input, archive_klass):
    archive_klass.convert(input)  # test generic archive
    # Create qualified archive that can be reversed
    compressed_type = Directory if input.is_dir else PlainText
    archive = archive_klass[compressed_type].convert(input)
    assert isinstance(archive, archive_klass)
    output = compressed_type.convert(archive)
    if isinstance(input, File):
        assert filecmp.cmp(output.fspath, input.fspath)
    else:
        _assert_extracted_dir_matches(output, input)


def _assert_extracted_dir_matches(extracted: Directory, reference: Directory):
    def assert_exact_match(cmp):
        assert (
            not cmp.left_only
        ), f"{cmp.left_only} missing from unarchved dir {cmp.right}"
        assert not cmp.right_only, (
            f"Additional {cmp.right_only} found in unarchived dir " + cmp.right
        )
        for subdir in cmp.subdirs.values():
            assert_exact_match(subdir)

    assert_exact_match(filecmp.dircmp(reference.fspath, extracted.fspath))



from __future__ import annotations
import pytest
import pydra.mark
from fileformats.core import from_mime, DataType, FileSet
from fileformats.core.mark import converter
from fileformats.archive import Zip
from fileformats.generic import DirectoryContaining
from fileformats.field import Array, Integer, Decimal, Text, Boolean
from fileformats.core.exceptions import (
    FileFormatsError,
    FormatConversionError,
    FormatRecognitionError,
    FormatMismatchError,
)
from fileformats.testing import (
    A,
    B,
    C,
    D,
    E,
    F,
    G,
    H,
    J,
    K,
    L,
    M,
    N,
    P,
    Q,
    R,
    TestField,
)


SpecificDataType = DataType.type_var("SpecificDataType")
SpecificFileSet = FileSet.type_var("SpecificFileSet")


def test_qualified_equivalence():

    assert F is F
    assert F[A] is F[A]
    assert F[A, B] is F[B, A]
    assert K[A, B] is not K[B, A]  # ordered qualifers
    assert F[A] is not F[B]
    assert F[SpecificDataType] is F[SpecificDataType]
    assert F[SpecificDataType] is not F[A]
    assert F[SpecificDataType] is not F[SpecificFileSet]


def test_subtype_testing():

    assert G.issubtype(F)
    assert not F.issubtype(G)
    assert J.issubtype(J)
    assert J.issubtype(H)
    assert J[A].issubtype(J)
    assert not H[A].issubtype(F)
    assert not H[A].issubtype(F[A])
    assert J[A, B].issubtype(J[A])
    assert not J[B].issubtype(J[A])
    assert not J[A].issubtype(J[A, B])
    assert J[A].issubtype(H[A])
    assert J[A, B, C].issubtype(H[A, B])
    assert not J[A].issubtype(H[B])
    assert FileSet.issubtype(DataType)
    assert SpecificFileSet.issubtype(SpecificDataType)
    assert F[SpecificFileSet].issubtype(F[SpecificDataType])
    assert L[A, E].issubtype(L[A, C])
    assert not L[E, A].issubtype(L[A, C])
    assert K[A, E, B].issubtype(K[A, C, B])
    assert N[J, K].issubtype(N[J, H])  # J is subclass of H,


def test_qualifier_fails():

    H[A, B, C]  # A, B, C are all allowable qualifier
    F[D]  # F has no restriction on qualifier types

    with pytest.raises(FileFormatsError) as e:
        H[D]
    assert "Invalid content types provided to" in str(e)

    with pytest.raises(FileFormatsError) as e:
        H[A, B, A]
    assert "Cannot have more than one occurrence of a qualifier" in str(e)

    K[A, B, A]  # ordered qualifiers allow repeats

    with pytest.raises(FileFormatsError) as e:
        Q[A]
    assert (
        "Default value for qualifiers attribute 'new_qualifiers_attr' needs to be set"
        in str(e)
    )

    with pytest.raises(FileFormatsError) as e:
        M[A, B]
    assert "Multiple qualifiers not permitted for " in str(e)


# (source_format=F, target_format=H)


@converter
@converter(source_format=F[A], target_format=H[A])  # Additional converter
@pydra.mark.task
def f2h(in_file: F) -> H:
    return in_file


def test_qualifier_converters():

    H.get_converter(F)
    assert F.get_converter(G) is None  # G is subtype of F
    with pytest.raises(FormatConversionError):  # Cannot convert to more specific type
        G.get_converter(F)
    H[A].get_converter(F[A])
    with pytest.raises(FormatConversionError):
        H[B].get_converter(F[B])
    assert F[A].get_converter(G[A]) is None
    assert F.get_converter(G[A]) is None
    with pytest.raises(FormatConversionError):
        F[A].get_converter(F[E])
    with pytest.raises(FormatConversionError):
        F[A].get_converter(G)


@converter(source_format=K[A, B], target_format=L[A, B])
@converter(source_format=K[C, D], target_format=L[D, C])
@converter(source_format=K[A, B, C], target_format=L[A, B])
@converter(source_format=K[A, C, B], target_format=L[C, A])
@pydra.mark.task
def k2l(in_file: K) -> L:
    return in_file


def test_ordered_qualifier_converters():
    L[A, B].get_converter(K[A, B])
    with pytest.raises(FormatConversionError):
        L[B, A].get_converter(K[A, B])
    L[D, C].get_converter(K[C, D])
    with pytest.raises(FormatConversionError):
        L[C, D].get_converter(K[C, D])

    L[C, A].get_converter(K[A, C, B])
    L[C, A].get_converter(K[A, E, B])  # E is a subtype of C


def test_mime_rountrips():

    assert DirectoryContaining[F].mime_like == "testing/f+directory-containing"
    assert from_mime("testing/f+directory-containing") is DirectoryContaining[F]

    # Directory is unordered so sort qualifiers to create unique mime
    assert DirectoryContaining[H, F].mime_like == "testing/f.h+directory-containing"
    assert from_mime("testing/f.h+directory-containing") is DirectoryContaining[F, H]

    # K is ordered
    assert K[B, A].mime_like == "testing/b.a+k"
    assert from_mime("testing/b.a+k") is K[B, A]
    assert from_mime("testing/b.a+k") is not K[A, B]

    with pytest.raises(FormatRecognitionError) as e:
        Array[TestField].mime_like
    assert "Cannot create reversible MIME type for " in str(e)


def test_inherited_qualifiers():

    assert Zip[G].mime_like == "testing/g+zip"
    assert from_mime("testing/g+zip") is Zip[G]


def test_arrays():

    Array[Integer]([1, 2, 3, 4])
    with pytest.raises(FormatMismatchError) as e:
        Array[Integer]([1.5, 2.2])
    assert "Cannot convert float (1.5) to integer field" in str(e)

    assert list(Array[Decimal]([1.5, 2.2])) == [1.5, 2.2]
    assert list(Array[Decimal](["1.5", "2.2"])) == [1.5, 2.2]
    assert list(Array[Decimal]("1.5, 2.2")) == [1.5, 2.2]
    assert list(Array[Decimal]("[1.5, 2.2]")) == [1.5, 2.2]
    assert list(Array[Text]("[1.5, 2.2]")) == ["1.5", "2.2"]

    assert list(Array[Boolean]("yes, YES, no, 0, 1, False, True, true")) == [
        True,
        True,
        False,
        False,
        True,
        False,
        True,
        True,
    ]

    assert list(from_mime("field/integer+array")("1,2,3,4,5")) == [1, 2, 3, 4, 5]


# Template-from-template conversions


@converter
@pydra.mark.task
def f2n_template(in_file: F[SpecificDataType]) -> N[SpecificDataType]:
    return in_file


@converter
@pydra.mark.task
def f2p_template(in_file: F[SpecificDataType]) -> P[SpecificDataType]:
    return in_file


@converter
@pydra.mark.task
def p2n_template(in_file: P[SpecificDataType]) -> N[SpecificDataType]:
    return in_file


def test_wildcard_template_from_template_conversion():
    H[A].get_converter(F[A])
    with pytest.raises(FormatConversionError):
        H[B].get_converter(F[A])


# Template from generic type to template


@converter
@pydra.mark.task
def generic2f(in_file: SpecificDataType) -> F[SpecificDataType]:
    return in_file


@converter
@pydra.mark.task
def generic2n(in_file: SpecificDataType) -> N[SpecificDataType, H]:
    return in_file


def test_wildcard_template_from_generic_conversion():

    F[J].get_converter(J)
    with pytest.raises(FormatConversionError):
        F[K].get_converter(J)

    N[J].get_converter(J)
    N[J, H].get_converter(J)
    with pytest.raises(FormatConversionError):
        F[K, H].get_converter(J)
    with pytest.raises(FormatConversionError):
        F[J, K].get_converter(J)


# Generic from template to  type


@converter
@pydra.mark.task
def f2generic(in_file: F[SpecificDataType]) -> SpecificDataType:
    return in_file


@converter
@pydra.mark.task
def n2generic(in_file: N[SpecificDataType, G]) -> SpecificDataType:
    return in_file


def test_wildcard_generic_from_template_conversion():
    J.get_converter(F[J])
    with pytest.raises(FormatConversionError):
        J.get_converter(F[K])


def test_wildcard_generic_from_multi_template_conversion():

    J.get_converter(N[J, G])
    with pytest.raises(FormatConversionError):
        J.get_converter(N[J, K])


@converter
@pydra.mark.task
def l2r(in_file: L[A, SpecificDataType, C]) -> R[A, SpecificDataType, C, D]:
    return in_file


def test_wildcard_ordered_qualifier_converters():
    R[A, B, C, D].get_converter(L[A, B, C])
    with pytest.raises(FormatConversionError):
        R[A, E, C, D].get_converter(L[A, B, C])



from pathlib import Path
import tempfile
import pydra.mark
import pydra.engine.specs
from fileformats.core import mark
from .raster import RasterImage, Bitmap, Gif, Jpeg, Png, Tiff


@mark.converter(target_format=Bitmap, output_format=Bitmap)
@mark.converter(target_format=Gif, output_format=Gif)
@mark.converter(target_format=Jpeg, output_format=Jpeg)
@mark.converter(target_format=Png, output_format=Png)
@mark.converter(target_format=Tiff, output_format=Tiff)
@pydra.mark.task
@pydra.mark.annotate({"return": {"out_file": RasterImage}})
def convert_image(in_file: RasterImage, output_format: type, out_dir: Path = None):
    data_array = in_file.load()
    if out_dir is None:
        out_dir = Path(tempfile.mkdtemp())
    output_path = out_dir / (in_file.fspath.stem + output_format.ext)
    return output_format.save_new(output_path, data_array)



import os
from copy import copy
from inspect import isclass
from warnings import warn
import traceback
import typing as ty
import importlib
import shutil
from operator import itemgetter
import itertools
from pathlib import Path
import hashlib
import logging
import attrs
from .utils import (
    subpackages,
    classproperty,
    fspaths_converter,
    to_mime_format_name,
    from_mime_format_name,
    STANDARD_NAMESPACES,
    hash_file,
    hash_dir,
    add_exc_note,
    describe_task,
)
from .converter import SubtypeVar
from .exceptions import (
    FileFormatsError,
    FormatMismatchError,
    FormatConversionError,
    FormatRecognitionError,
)

# Tools imported from Arcana, will remove again once file-formats and "cells"
# have been split
REQUIRED_ANNOTATION = "__fileformats_required__"
CHECK_ANNOTATION = "__fileformats_check__"


logger = logging.getLogger("fileformats")


class DataType:
    is_fileset = False
    is_field = False

    @classmethod
    def type_var(cls, name):
        return SubtypeVar(name, cls)

    @classmethod
    def matches(cls, values) -> bool:
        """Checks whether the given value (fspaths for file-sets) match the datatype
        specified by the class

        Parameters
        ----------
        values : ty.Any
            values to check whether they match the given datatype

        Returns
        -------
        matches : bool
            whether the datatype matches the provided values
        """
        try:
            cls(values)
        except FormatMismatchError:
            return False
        else:
            return True

    @classmethod
    def issubtype(cls, super_type: type, allow_same: bool = True):
        """Check to see whether datatype class is a subtype of a given super class.
        In this case the subtype is expected to be able to be treated as if it was
        the super class.

        Overridden in the ``WithQualifiers`` mixin to add support for
        qualified subtypes

        Parameters
        ----------
        super_type : type
            the class to check whether the given class is a subtype of
        allow_same : bool, optional
            whether there is a match if the classes are the same, by default True

        Returns
        -------
        is_subtype : bool
            whether or not the current class can be considered a subtype of the super (or
            is the super itself)
        """
        if allow_same and cls is super_type:
            return True
        if isinstance(super_type, SubtypeVar):
            super_type = super_type.base
        return issubclass(cls, super_type)

    @classproperty
    def namespace(cls):
        """The "namespace" the format belongs to under the "fileformats" umbrella
        namespace"""
        module_parts = cls.__module__.split(".")
        if module_parts[0] != "fileformats":
            raise FileFormatsError(
                f"Cannot create reversible MIME type for {cls} as it is not in the "
                "fileformats namespace"
            )
        return module_parts[1]

    @classproperty
    def all_types(self):
        return itertools.chain(FileSet.all_formats, Field.all_fields)

    @classmethod
    def subclasses(cls):
        """Iterate over all installed subclasses"""
        for subpkg in subpackages():
            for attr_name in dir(subpkg):
                attr = getattr(subpkg, attr_name)
                if isclass(attr) and issubclass(attr, cls):
                    yield attr

    @classproperty
    def mime_like(cls):
        """Generates a "MIME-like" identifier from a format class (i.e.
        an identifier for a non-MIME class in the MIME style), e.g.

            fileformats.text.Plain to "text/plain"

        and

            fileformats.image.TiffFx to "image/tiff-fx"

        Parameters
        ----------
        klass : type(FileSet)
            FileSet subclass
        iana_mime : bool
            whether to use standardised IANA format or a more relaxed type format corresponding
            to the fileformats extension the type belongs to

        Returns
        -------
        type
            the corresponding file format class
        """
        mime = f"{cls.namespace}/{to_mime_format_name(cls.__name__)}"
        try:
            cls.from_mime(mime)
        except FormatRecognitionError as e:
            add_exc_note(
                e,
                (
                    f"Cannot create reversible MIME type for {cls} as it is not present "
                    f"in a top-level fileformats namespace package '{cls.namespace}'"
                ),
            )
            raise e
        return mime

    @classmethod
    def from_mime(cls, mime_string):
        """Resolves a FileFormat class from a MIME (IANA) or "MIME-like" identifier (i.e.
        an identifier for a non-MIME class in the MIME style), e.g.

            "text/plain" resolves to fileformats.text.Plain

        and

            "image/tiff-fx" resolves to fileformats.image.TiffFx

        Parameters
        ----------
        mime_string : str
            MIME identifier

        Returns
        -------
        type
            the corresponding file format class
        """
        namespace, format_name = mime_string.split("/")
        try:
            return FileSet.formats_by_iana_mime[mime_string]
        except KeyError:
            pass
        if namespace == "application":
            # We treat the "application" namespace as a catch-all for any formats that are
            # not explicitly covered by the IANA standard (which is kind of how the IANA
            # treats it). Therefore, we loop through all subclasses across the different
            # namespaces to find one that matches the name.
            if not format_name.startswith("x-"):
                raise FormatRecognitionError(
                    "Did not find class matching official (i.e. non-extension) MIME type "
                    f"{mime_string} (i.e. one not starting with 'application/x-'"
                ) from None
            format_name = format_name[2:]  # remove "x-" prefix
            matching_name = FileSet.formats_by_name[format_name]
            if not matching_name:
                namespace_names = [n.__name__ for n in subpackages()]
                class_name = from_mime_format_name(format_name)
                raise FormatRecognitionError(
                    f"Did not find class matching extension the class name '{class_name}' "
                    f"corresponding to MIME type '{mime_string}' "
                    f"in any of the installed namespaces: {namespace_names}"
                )
            elif len(matching_name) > 1:
                namespace_names = [f.__module__.__name__ for f in matching_name]
                raise FormatRecognitionError(
                    f"Ambiguous extended MIME type '{mime_string}', could refer to "
                    f"{', '.join(repr(f) for f in matching_name)} installed types. "
                    f"Explicitly set the 'iana_mime' attribute on one or all of these types "
                    f"to disambiguate, or uninstall all but one of the following "
                    "namespaces: "
                )
            else:
                klass = next(iter(matching_name))
        else:
            class_name = from_mime_format_name(format_name)
            try:
                module = importlib.import_module("fileformats." + namespace)
            except ImportError:
                raise FormatRecognitionError(
                    f"Did not find fileformats namespace package corresponding to {namespace} "
                    f"required to interpret '{mime_string}' MIME, or MIME-like, type. "
                    f"try installing the namespace package with "
                    f"'python3 -m pip install fileformats-{namespace}'."
                ) from None
            try:
                klass = getattr(module, class_name)
            except AttributeError:
                if "+" in format_name:
                    qualifier_names, qualified_name = format_name.split("+")
                    try:
                        qualifiers = [
                            getattr(module, from_mime_format_name(q))
                            for q in qualifier_names.split(".")
                        ]
                    except AttributeError:
                        raise FormatRecognitionError(
                            f"Could not load qualifiers [{qualifier_names}] from "
                            f"fileformats.{namespace}, corresponding to MIME, "
                            f"or MIME-like, type {mime_string}"
                        ) from None
                    try:
                        qualified = getattr(
                            module, from_mime_format_name(qualified_name)
                        )
                    except AttributeError:
                        try:
                            qualified = cls.generically_qualifies_by_name[
                                qualified_name
                            ]
                        except KeyError:
                            raise FormatRecognitionError(
                                f"Could not load qualified class '{qualified_name}' from "
                                f"fileformats.{namespace} or list of generic types "
                                f"({list(cls.generically_qualifies_by_name)}), "
                                f"corresponding to MIME, or MIME-like, type {mime_string}"
                            ) from None
                    klass = qualified[qualifiers]
                else:
                    raise FormatRecognitionError(
                        f"Did not find '{class_name}' class in fileformats.{namespace} "
                        f"corresponding to MIME, or MIME-like, type {mime_string}"
                    ) from None
        return klass

    @classproperty
    def generically_qualifies_by_name(cls):
        if cls._generically_qualifies_by_name is None:
            cls._generically_qualifies_by_name = {
                to_mime_format_name(f.__name__): f
                for f in FileSet.all_formats
                if getattr(f, "generically_qualifies", False)
            }
        return cls._generically_qualifies_by_name

    _generically_qualifies_by_name = None  # Register all generically qualified types


@attrs.define
class FileSet(DataType):
    """
    The base class for all format types within the fileformats package. A generic
    representation of a collection of files related to a single data resource. A
    file-set can be a single file or directory or a collection thereof, such as a
    primary file with a "side-car" header.

    Parameters
    ----------
    fspaths : set[Path]
        a set of file-system paths pointing to all the resources in the file-set
    checks : bool
        whether to run in-depth "checks" to verify the file format
    """

    fspaths: ty.FrozenSet[Path] = attrs.field(default=None, converter=fspaths_converter)
    _metadata: ty.Dict[str, ty.Any] = attrs.field(
        default=None,
        init=False,
        repr=False,
        eq=False,
        hash=False,
    )

    # Explicitly set the Internet Assigned Numbers Authority (https://iana_mime.org) MIME
    # type to None for any base classes that should not correspond to a MIME or MIME-like
    # type.
    iana_mime = None
    ext = ""

    # Store converters registered by @converter decorator that convert to FileSet
    # NB: each class will have its own version of this dictionary
    converters = {}

    is_fileset = True

    def __attrs_post_init__(self):
        # Check required properties don't raise errors
        for prop_name in self.required_properties():
            getattr(self, prop_name)
        # Loop through all attributes and find methods marked by CHECK_ANNOTATION
        for check in self.checks():
            getattr(self, check)()

    @fspaths.validator
    def validate_fspaths(self, _, fspaths):
        if not fspaths:
            raise FileFormatsError(f"No file-system paths provided to {self}")
        missing = [p for p in fspaths if not p or not p.exists()]
        if missing:
            missing_str = "\n".join(str(p) for p in missing)
            msg = (
                f"The following file system paths provided to {type(self)} do not "
                f"exist:\n\n{missing_str}\n\n"
            )
            present_parents = set()
            for fspath in missing:
                if fspath:
                    if fspath.parent.exists():
                        present_parents.add(fspath.parent)
            for parent in present_parents:
                msg += (
                    f"\n\nFiles in the present parent directory '{str(parent)}' are:\n"
                )
                msg += "\n".join(str(p) for p in parent.iterdir())
            raise FileNotFoundError(msg)

    @property
    def metadata(self):
        if self._metadata is None and hasattr(self, "load_metadata"):
            self._metadata = self.load_metadata()
        return self._metadata

    @classproperty
    def mime_type(cls):
        """Generates a MIME type (IANA) identifier from a format class. If an official
        IANA MIME type doesn't exist it will create one in the in the MIME style, e.g.

            fileformats.text.Plain to "text/plain"

            fileformats.image.TiffFx to "image/tiff-fx"

            fileformats.mynamespace.MyFormat to "application/x-my-format

        Returns
        -------
        str
            the MIME type corresponding to the class
        """
        if getattr(cls, "iana_mime", None) is not None:
            mime_type = cls.iana_mime
        else:
            format_name = to_mime_format_name(cls.__name__)
            mime_type = f"application/x-{format_name}"
        return mime_type

    @classmethod
    def required_properties(cls):
        """Find all properties required to treat file-set as being in the format specified
        by the class

        Returns
        -------
        iter(str)
            an iterator over all properties names marked as "required"
        """
        fileset_props = dir(FileSet)
        for attr_name in dir(cls):
            if attr_name in fileset_props:
                continue
            klass_attr = getattr(cls, attr_name)
            if isinstance(klass_attr, property):
                try:
                    klass_attr.fget.__annotations__[REQUIRED_ANNOTATION]
                except KeyError:
                    pass
                else:
                    yield attr_name

    def required_paths(self) -> ty.Set[Path]:
        """Returns all fspaths that are required for the format"""
        required = set()
        for prop_name in self.required_properties():
            prop = getattr(self, prop_name)
            paths = []
            if hasattr(prop, "required_paths"):
                paths = prop.required_paths()
            elif isinstance(prop, os.PathLike):
                paths = [Path(prop)]
            elif isinstance(prop, ty.Iterable):
                for p in prop:
                    if isinstance(p, os.PathLike):
                        paths.append(Path(p))
            for path in paths:
                if path in self.fspaths:
                    required.add(path)
        return required

    def trim_paths(self):
        """Trims paths in fspaths to only those that are "required" by the format class
        i.e. returned by a required property"""
        self.fspaths = fspaths_converter(self.required_paths())
        return self

    @classmethod
    def checks(cls):
        """Find all methods used to check the validity of the file format

        Returns
        -------
        iter(str)
            an iterator over all method names marked as a "check"
        """
        # Loop through all attributes and find methods marked by CHECK_ANNOTATION
        fileset_props = dir(FileSet)
        for attr_name in dir(cls):
            if attr_name in fileset_props:
                continue
            klass_attr = getattr(cls, attr_name)
            try:
                klass_attr.__annotations__[CHECK_ANNOTATION]
            except (AttributeError, KeyError):
                pass
            else:
                yield attr_name

    def copy_to(
        self,
        dest_dir: Path,
        stem: str = None,
        symlink: bool = False,
        trim: bool = True,
        make_dirs: bool = False,
    ):
        """Copies the file-set to a new directory, optionally renaming the files
        to have consistent name-stems.

        Parameters
        ----------
        dest_dir : str
            Path to the parent directory to save the file-set
        stem: str, optional
            the file name excluding file extensions, to give the files/dirs in the parent
            directory, by default the original file name is used
        symlink : bool, optional
            Use symbolic links instead of copying files to new location, false by default
        trim : bool, optional
            Only copy the paths in the file-set that are "required" by the format, true by default
        make_dirs : bool, optional
            Make the parent destination and all missing ancestors if they are missing, false by default
        """
        dest_dir = Path(dest_dir)  # ensure a Path not a string
        if make_dirs:
            dest_dir.mkdir(parents=True, exist_ok=True)
        if symlink:
            copy_dir = copy_file = os.symlink
        else:
            copy_dir = shutil.copytree
            copy_file = shutil.copyfile
        new_paths = []
        if trim:
            fspaths_to_copy = self.required_paths()
        else:
            fspaths_to_copy = self.fspaths
        for fspath in fspaths_to_copy:
            if stem:
                new_fname = stem + "".join(fspath.suffixes)
            else:
                new_fname = fspath.name
            new_path = dest_dir / new_fname
            if fspath.is_dir():
                copy_dir(fspath, new_path)
            else:
                copy_file(fspath, new_path)
            new_paths.append(new_path)
        return type(self)(new_paths)

    def select_by_ext(self, fileformat: type = None, allow_none: bool = False) -> Path:
        """Selects a single path from a set of file-system paths based on the file
        extension

        Parameters
        ----------
        fileformat : type
            the format class of the path to select
        allow_none : bool
            whether to return None instead of raising an error if the file is not found

        Returns
        -------
        Path or None
            the selected file-system path that matches the extension or None if not found
            and `allow_none` is True

        Raises
        ------
        FileFormatError
            if no paths match the extension and `allow_none` is False
        FileFormatError
            if more than one paths matches the extension
        """
        if fileformat is None:
            fileformat = type(self)
        exts = [fileformat.ext]
        try:
            exts.extend(fileformat.alternate_exts)
        except AttributeError:
            pass
        matches = self.matching_exts(self.fspaths, exts)
        if not matches:
            if allow_none:
                return None
            paths_str = ", ".join(str(p) for p in self.fspaths)
            raise FormatMismatchError(
                f"No matching files with extensions in {exts} in file set {paths_str}"
            )
        elif len(matches) > 1:
            matches_str = ", ".join(str(p) for p in matches)
            raise FormatMismatchError(
                f"Multiple files with {exts} extensions found in : {matches_str}"
            )
        return matches[0]

    @classmethod
    def matching_exts(cls, fspaths: ty.Set[Path], exts: ty.List[str]) -> ty.List[Path]:
        """Returns the paths out of the candidates provided that matches the
        given extension (by default the extension of the class)

        Parameters
        ----------
        fspaths: list[Path]
            The paths to select from
        ext: list[str]
            the extensions to match

        Returns
        -------
        list[Path]
            the matching paths

        Raises
        ------
        FileFormatError
            When no paths match or more than one path matches the given extension"""
        return [p for p in fspaths if any(str(p).endswith(e) for e in exts)]

    @classmethod
    def convert(cls, fileset, plugin="serial", task_name=None, **kwargs):
        """Convert a given file-set into the format specified by the class

        Parameters
        ----------
        fileset : FileSet
            the file-set object to convert
        plugin : str
            the "execution plugin" used to run the conversion task
        task_name : str
            the name given to the converter task
        **kwargs
            args to pass to the conversion process

        Returns
        -------
        FileSet
            the file-set converted into the type of the current class
        """
        if task_name is None:
            task_name = f"{type(fileset).__name__}_to_{cls.__name__}_{id(fileset)}"
        # Make unique, yet somewhat recognisable task name
        task = cls.get_converter(source_format=type(fileset), name=task_name, **kwargs)
        result = task(in_file=fileset, plugin=plugin)
        out_file = result.output.out_file
        if not isinstance(out_file, cls):
            out_file = cls(out_file)
        return out_file

    @classmethod
    def get_converter(cls, source_format: type, name: str = "converter", **kwargs):
        """Get a converter that converts from the source format type
        into the format specified by the class

        Parameters
        ----------
        source_format : type
            the format to convert from
        name : str
            the name given to the converter task
        **kwargs
            passed on to the task init method to customise the conversion

        Returns
        -------
        pydra.engine.TaskBase or None
            a pydra task or workflow that performs the conversion, or None if no
            conversion is required

        Raises
        ------
        FileFormatConversionError
            no converters found between source and dest format
        FileFormatConversionError
            ambiguous (i.e. more than one) converters found between source and dest format
        """
        if source_format.issubtype(cls):
            return None
        converters = (
            cls.get_converters_dict()
        )  # triggers loading of standard converters for target format
        # Ensure standard converters from source format are loaded
        source_format.import_standard_converters()
        try:
            unqualified = source_format.unqualified
        except AttributeError:
            pass
        else:
            unqualified.import_standard_converters()
        try:
            converter_tuple = converters[source_format]
        except KeyError:
            # If no direct mapping check for mapping from source super types and wildcard
            # matches
            available_converters = cls.get_converter_tuples(source_format)
            if len(available_converters) > 1:
                available_str = "\n".join(
                    describe_task(a[0]) for a in available_converters
                )
                raise FormatConversionError(
                    f"Ambiguous converters found between '{cls.mime_like}' and "
                    f"'{source_format.mime_like}':\n{available_str}"
                ) from None
            elif not available_converters:
                raise FormatConversionError(
                    f"Could not find converter between '{source_format.mime_like}' and "
                    f"'{cls.mime_like}' formats"
                ) from None
            converter_tuple = available_converters[0]
            # Store mapping for future reference
            converters[source_format] = converter_tuple
        converter, conv_kwargs = converter_tuple
        if kwargs:
            # Merge kwargs provided to get_conveter with stored kwargs
            conv_kwargs = copy(conv_kwargs)
            conv_kwargs.update(kwargs)
        return converter(name=name, **conv_kwargs)

    @classmethod
    def get_converters_dict(cls, klass=None):
        # Only access converters to the specific class, not superclasses (which may not
        # be able to convert to the specific type)
        if klass is None:
            klass = cls
        try:
            converters_dict = klass.__dict__["converters"]
        except KeyError:
            converters_dict = {}
            klass.converters = converters_dict
            klass.import_standard_converters()
        return converters_dict

    @classmethod
    def import_standard_converters(cls):
        """Attempts to import standard converters for the format class, which are
        located at `fileformats.<namespace>.converters`
        """
        standard_converters_module = f"fileformats.{cls.namespace}.converters"
        try:
            importlib.import_module(standard_converters_module)
        except ImportError as e:
            if str(e) != f"No module named '{standard_converters_module}'":
                if cls.namespace in STANDARD_NAMESPACES:
                    pkg = "fileformats"
                else:
                    pkg = f"fileformats-{cls.namespace}"
                warn(
                    f"Could not import standard converters for '{cls.namespace}' namespace, "
                    f"please install '{pkg}' with the 'extended' install option to "
                    f"use converters for {cls.namespace}, i.e.\n\n"
                    f"    $ python3 -m pip install '{pkg}[extended]':\n\n"
                    f"Import error was:\n{traceback.format_exc()}"
                )

    @classmethod
    def get_converter_tuples(
        cls, source_format: type
    ) -> ty.List[ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]]]:
        """Search the registered converters to find any matches and return list of
        task and associated key-word args to perform the conversion between source and
        target formats

        Parameters
        ----------
        source_format : type(FileSet)
            the source format to convert from

        Returns
        -------
        available : list[tuple[TaskBase, dict[str, Any]]]
            list of available converters between the source and target formats
        """
        converters_dict = cls.get_converters_dict()
        available = []
        for src_frmt, converter in converters_dict.items():
            if len(converter) == 2:  # Ignore converters with wildcards at this point
                if source_format.issubtype(src_frmt):
                    available.append(converter)
        if not available and hasattr(source_format, "unqualified"):
            available = SubtypeVar.get_converter_tuples(
                source_format, target_format=cls
            )
        return available

    @classmethod
    def register_converter(
        cls,
        source_format: type,
        converter_tuple: ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]],
    ):
        """Registers a converter task within a class attribute. Called by the @fileformats.mark.converter
        decorator.

        Parameters
        ----------
        source_format : type
            the source format to register a converter from
        task_spec : ty.Callable
            a callable that resolves to a Pydra task
        converter_kwargs : dict
            additional keyword arguments to be passed to the task spec at initialisation
            time

        Raises
        ------
        FormatConversionError
            if there is already a converter registered between the two types
        """
        converters_dict = cls.get_converters_dict()
        # If no converters are loaded, attempt to load from the standard location
        if source_format in converters_dict:
            prev = cls.converters[source_format][0]
            raise FormatConversionError(
                f"There is already a converter registered between {source_format.__name__} "
                f"and {cls.__name__}: {describe_task(prev)}"
            )
        converters_dict[source_format] = converter_tuple

    @classproperty
    def all_formats(cls):
        """Iterate over all FileSet formats in fileformats.* namespaces"""
        if cls._all_formats is None:
            cls._all_formats = set(
                f for f in FileSet.subclasses() if f.__dict__.get("iana_mime", True)
            )
        return cls._all_formats

    @classproperty
    def standard_formats(cls):
        """Iterate over all formats in the standard fileformats.* namespaces"""
        return (f for f in cls.all_formats if f.namespace in STANDARD_NAMESPACES)

    @classproperty
    def formats_by_iana_mime(cls):
        """a dictionary containing all formats by their IANA MIME type (if applicable)"""
        if cls._formats_by_iana_mime is None:
            cls._formats_by_iana_mime = {
                f.iana_mime: f
                for f in FileSet.all_formats
                if getattr(f, "iana_mime", None) is not None
            }
        return cls._formats_by_iana_mime

    @classproperty
    def formats_by_name(cls):
        """a dictionary containing lists of formats by their translated class names,
        i.e. their can be more than one format with the same translated name"""
        if cls._formats_by_name is None:
            cls._formats_by_name = {
                k: set(v for _, v in g)
                for k, g in itertools.groupby(
                    sorted(
                        (
                            (to_mime_format_name(f.__name__), f)
                            for f in FileSet.all_formats
                            if getattr(f, "iana_mime", None) is None
                        ),
                        key=itemgetter(0),
                    ),
                    key=itemgetter(0),
                )
            }
        return cls._formats_by_name

    def hash_files(
        self,
        crypto=None,
        chunk_len=8192,
        relative_to: os.PathLike = None,
        **kwargs,
    ):
        """Calculate hashes for all files within the file set within a dictionary.
        Hashes for files within directories are nested within separate dictionaries
        for each (sub)directory.

        Parameters
        ----------
        crypto : function, optional
            the cryptography method used to hash the files, by default hashlib.sha256
        chunk_len : int, optional
            the chunk length to break up the file and calculate the hash over, by default 8192
        relative_to : Path, optional
            the base path by which to record the file system paths in the dictionary keys
            to, by default None
        **kwargs
            keyword args passed directly through to the ``hash_dir`` function

        Returns
        -------
        file_hashes : dict[str, str]
            hashes for all files in the file-set, addressed by their directory relative
            to the ``relative_to`` path
        """
        if relative_to is None:
            relative_to = Path(os.path.commonpath(self.fspaths))
            if all(p.is_file() and p.parent == relative_to for p in self.fspaths):
                relative_to /= os.path.commonprefix(
                    [p.name for p in self.fspaths]
                ).rstrip(".")
        relative_to = str(relative_to)
        if Path(relative_to).is_dir() and not relative_to.endswith(os.path.sep):
            relative_to += os.path.sep
        if crypto is None:
            crypto = hashlib.sha256

        file_hashes = {}
        for key, fspath in sorted(
            ((str(p)[len(relative_to) :], p) for p in self.fspaths),
            key=itemgetter(0),
        ):
            if fspath.is_file():
                file_hashes[key] = hash_file(fspath, chunk_len=chunk_len, crypto=crypto)
            elif fspath.is_dir():
                file_hashes.update(
                    hash_dir(
                        fspath,
                        chunk_len=chunk_len,
                        crypto=crypto,
                        relative_to=relative_to,
                        **kwargs,
                    )
                )
            else:
                raise RuntimeError(f"Cannot hash {self} as {fspath} no longer exists")
        return file_hashes

    def hash(self, crypto=None, *args, **kwargs):
        """Calculate a unique hash for the file-set based on the relative paths and
        contents of its constituent files

        Parameters
        ----------
        crypto : function, optional
            the cryptography method used to hash the files, by default hashlib.sha256
        chunk_len : int, optional
            the chunk length to break up the file and calculate the hash over, by default 8192
        relative_to : Path, optional
            the base path by which to record the file system paths in the dictionary keys
            to, by default None
        **kwargs
            keyword args passed directly through to the ``hash_dir`` function

        Returns
        -------
        hash : str
            unique hash for the file-set
        """
        if crypto is None:
            crypto = hashlib.sha256
        crytpo_obj = crypto()
        for path, hash in self.hash_files(crypto=crypto, *args, **kwargs).items():
            crytpo_obj.update(path.encode())
            crytpo_obj.update(hash.encode())
        return crytpo_obj.hexdigest()

    _all_formats = None
    _formats_by_iana_mime = None
    _formats_by_name = None


@attrs.define
class Field(DataType):
    value = attrs.field()

    type = None
    is_field = True

    def __str__(self):
        return str(self.value)

    @property
    def metadata(self):
        return {}

    @classproperty
    def all_fields(cls):
        """Iterate over all field formats in fileformats.* namespaces"""
        if cls._all_fields is None:
            cls._all_fields = [f for f in Field.subclasses() if f.type is not None]
        return cls._all_fields

    _all_fields = None



from pathlib import Path
import tempfile
import pydra.mark
import pydra.engine.specs
from fileformats.core import mark
from . import DataSerialization, Json, Yaml


@mark.converter(target_format=Json, output_format=Json)
@mark.converter(target_format=Yaml, output_format=Yaml)
@pydra.mark.task
@pydra.mark.annotate({"return": {"out_file": DataSerialization}})
def convert_data_serialization(
    in_file: DataSerialization, output_format: type, out_dir: Path = None
):
    dct = in_file.load()
    if out_dir is None:
        out_dir = Path(tempfile.mkdtemp())
    output_path = out_dir / (in_file.fspath.stem + output_format.ext)
    return output_format.save_new(output_path, dct)



import pytest
from fileformats.generic import File
from fileformats.core.exceptions import FileFormatsError, FormatMismatchError
from fileformats.core import mark
from conftest import write_test_file


def test_init_args(work_dir):
    fspath = work_dir / "test.txt"
    write_test_file(fspath)
    File([fspath])
    with pytest.raises(TypeError):
        File(fspath, fspath)


# def test_metadata_iterator_fail(work_dir):
#     fspath = work_dir / "test.txt"
#     write_test_file(fspath)
#     file = File(fspath)
#     with pytest.raises(NotImplementedError):
#         iter(file.metadata)


class TestFile(File):

    ext = ".tst"


def test_multiple_matches1(work_dir):
    fspath1 = work_dir / "test1.tst"
    write_test_file(fspath1)
    fspath2 = work_dir / "test2.tst"
    write_test_file(fspath2)
    with pytest.raises(FormatMismatchError):
        File([fspath1, fspath2])


def test_multiple_matches2(work_dir):
    fspath1 = work_dir / "test1.tst"
    write_test_file(fspath1)
    fspath2 = work_dir / "test2.tst"
    write_test_file(fspath2)
    with pytest.raises(FormatMismatchError):
        TestFile([fspath1, fspath2])


def test_missing_files(work_dir):
    fspath = work_dir / "test.tst"
    write_test_file(fspath)
    with pytest.raises(FileNotFoundError):
        TestFile([fspath, work_dir / "missing1.txt", work_dir / "missing2.txt"])


class ImageWithInlineHeader(File):

    ext = ".img"
    binary = True

    header_separator = b"---END HEADER---"

    def load_metadata(self):
        hdr = self.contents.split(self.header_separator)[0].decode("utf-8")
        return {k: int(v) for k, v in (ln.split(":") for ln in hdr.splitlines())}


@pytest.mark.xfail(reason="Disabled separate header metadata")
def test_header_overwrite(work_dir):

    fspath = work_dir / "image.img"
    hdr = {
        "x": 10,
        "y": 20,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    file = ImageWithInlineHeader(fspath, metadata={"x": "100"})
    with pytest.raises(FileFormatsError):
        file.metadata["y"]


class YFile(ImageWithInlineHeader):
    @mark.required
    @property
    def y(self):
        return self.metadata["y"]

    @mark.check
    def y_value(self):
        if self.y <= 10:
            raise FormatMismatchError(f"'y' property is not > 10 ({self.y})")


def test_required_check_op(work_dir):
    fspath = work_dir / "image.img"
    hdr = {
        "x": 10,
        "y": 20,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    assert YFile.matches(fspath)


def test_required_check_op_fail(work_dir):
    fspath = work_dir / "image.img"
    hdr = {
        "x": 10,
        "y": 10,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    assert not YFile.matches(fspath)



from ..core import __version__
from fileformats.generic import File
from fileformats.core.mixin import WithMagicNumber


class Audio(File):
    "Base class for audio file formats"
    binary = True
    iana_mime = None


# Compressed formats
class Mpeg(Audio):
    ext = ".mp3"
    alternate_exts = (".mp1", ".mp2")
    iana_mime = "audio/mpeg"


class Mp4(WithMagicNumber, Audio):
    ext = ".mp4"
    iana_mime = "audio/mp4"
    magic_number = "6674797069736F6D"


class Aac(Audio):
    ext = ".aac"
    alternate_exts = (".adts", ".loas", ".ass")
    iana_mime = "audio/aac"


class Wav(Audio):
    ext = ".wav"
    iana_mime = "audio/wav"
